//-----------------------------------------------------------------------------
// Copyright (c) 2007-2008 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------
//
// This D3DX Effect file implements tangent space normal mapping.
//
// Three techniques are provided for each of the three lighting types:
// directional lighting, point lighting, and spotlight lighting.
//
// Tangent space normal mapping involves evaluating the lighting equations per
// pixel in the surface local coordinate space of the normal map texture.
// Rather than using the mesh's interpolated vertex normals tangent space
// normal mapping uses the normals from the normal map texture instead. Doing
// this allows additional surface detail to be applied to each face of the mesh
// without requiring extra geometry to be added to each face.
//
// The lighting vectors are transformed into the normal map texture's tangent
// space using a rotation matrix. This rotation matrix must be calculated for
// each face of the mesh. Each face has its own rotation matrix because each
// face is uniquely texture mapped to the normal map texture and the rotation
// matrix must take this into consideration. The rotation matrix is constructed
// using each face's tangent, bitangent, and normal vectors.
//
// The pixel shader used in tangent space normal mapping is identical to the
// pixel shader used in per-pixel lighting. The only difference is the normal
// used is read from the normal map texture rather that being interpolated by
// the rasterizer. Prior to being stored in the normal map texture the normals
// are in the range [-1,1]. Once stored in the normal map texture the normals
// will be compressed to the range [0,255]. When the normals are accessed from
// the pixel shader HLSL will remap then to the range [0,1]. To convert the
// normals back to the original [-1,1] range we need to perform a scale and a
// bias. We do this by: tex2D(normalMap, IN.texCoord) * 2.0f - 1.0f.
//
// Light attenuation for the point and spot lighting models is based on a
// light radius. Light is at its brightest at the center of the sphere defined
// by the light radius. There is no lighting at the edges of this sphere.
//
//-----------------------------------------------------------------------------

struct Light
{
	float3 dir;				// world space direction
	float3 pos;				// world space position
	float4 ambient;
	float4 diffuse;
	float4 specular;
	float spotInnerCone;	// spot light inner cone (theta) angle
	float spotOuterCone;	// spot light outer cone (phi) angle
	float radius;           // applies to point and spot lights only
};

struct Material
{
	float4 ambient;
	float4 diffuse;
	float4 emissive;
	float4 specular;
	float shininess;
};

//-----------------------------------------------------------------------------
// Globals.
//-----------------------------------------------------------------------------

float4x4 g_mWorld;
float4x4 g_mWorldInverseTranspose;
float4x4 g_mWorldViewProjection;
float4x4 g_mViewInv;

//float3 cameraPos;
float4 globalAmbient = { 0.2f, 0.2f, 0.2f, 1.0f };
float4 blue = { 0.44f, 0.57f, 0.75f, 1.0f };
float4 orange = {0.974f, 0.338f, 0.034f, 1.0f};
float4 grey = {0.12f, 0.103f, 0.092f, 1.0f};
float4 white = {0.914f, 0.865, 0.719f, 1.0f};
float specPower;

bool   g_outNormals;
bool   g_mode;

Light light;
Material material;

//-----------------------------------------------------------------------------
// Textures.
//-----------------------------------------------------------------------------

texture g_diffuseTexture;
texture g_normalTexture;
texture g_envTexture;
texture g_irradTexture;

samplerCUBE irradMap = sampler_state
{
    Texture = <g_irradTexture>;
    MinFilter = Linear;
    MipFilter = Linear;
    MagFilter = Linear;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

samplerCUBE envMap = sampler_state
{
    Texture = <g_envTexture>;
    MinFilter = Linear;
    MipFilter = Linear;
    MagFilter = Linear;
    AddressU = Clamp;
    AddressV = Clamp;
    AddressW = Clamp;
};

sampler2D colorMap = sampler_state
{
	Texture = <g_diffuseTexture>;
    MagFilter = Anisotropic;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU = Wrap;
    AddressV = Wrap;
};

sampler2D normalMap = sampler_state
{
    Texture = <g_normalTexture>;
    MagFilter = Anisotropic;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU = Wrap;
    AddressV = Wrap;
};

//-----------------------------------------------------------------------------
// Vertex Shaders.
//-----------------------------------------------------------------------------

struct VS_INPUT
{
	float3 position : POSITION;
	float2 texCoord : TEXCOORD0;
	float3 normal : NORMAL;
    float3 tangent : TANGENT;  //4
};

struct VS_OUTPUT_DIR
{
	float4 position : POSITION;
	float2 texCoord : TEXCOORD0;
	float3 halfVector : TEXCOORD1;
	float3 lightDir : TEXCOORD2;
	float3 viewDir : TEXCOORD3;//////////////////
	float3 WorldNormal	: TEXCOORD4;///////
	float4 diffuse : COLOR0;
	float4 specular : COLOR1;
};

struct VS_OUTPUT_POINT
{
	float4 position : POSITION;
	float2 texCoord : TEXCOORD0;
	float3 viewDir : TEXCOORD1;
	float3 lightDir : TEXCOORD2;
	float3 WorldNormal	: TEXCOORD3;///////
	float4 diffuse : COLOR0;
	float4 specular : COLOR1;
};

struct VS_OUTPUT_SPOT
{
	float4 position : POSITION;
	float2 texCoord : TEXCOORD0;
	float3 viewDir : TEXCOORD1;
	float3 lightDir : TEXCOORD2;
	float3 spotDir : TEXCOORD3;
	float4 diffuse : COLOR0;
	float4 specular : COLOR1;
};

VS_OUTPUT_DIR VS_DirLighting(VS_INPUT IN)
{
	VS_OUTPUT_DIR OUT;

	float3 worldPos = mul(float4(IN.position, 1.0f), g_mWorld).xyz;
	float3 lightDir = -light.dir;
	float3 viewDir = g_mViewInv[3].xyz - worldPos;
	float3 halfVector = normalize(normalize(lightDir) + normalize(viewDir));
		
	float3 n = normalize(mul(IN.normal, (float3x3)g_mWorldInverseTranspose));//g_mWorldInverseTranspose
	float3 t = normalize(mul(IN.tangent, (float3x3)g_mWorldInverseTranspose));//.xyz  g_mWorldInverseTranspose
	float3 b = normalize(cross(t, n));// n, t * IN.tangent.w
	float3x3 tbnMatrix = float3x3(t.x, b.x, n.x,
	                              t.y, b.y, n.y,
	                              t.z, b.z, n.z);

	OUT.position = mul(float4(IN.position, 1.0f), g_mWorldViewProjection);
	OUT.texCoord = IN.texCoord;
	OUT.halfVector = mul(halfVector, tbnMatrix);
	OUT.lightDir = mul(lightDir, tbnMatrix);
	OUT.viewDir = mul(viewDir, tbnMatrix);//////////////////////////
	OUT.WorldNormal = mul(IN.normal, (float3x3)g_mWorld);//////////////////////////
	OUT.diffuse = material.diffuse * light.diffuse;
	OUT.specular = material.specular * light.specular;

	return OUT;
}

VS_OUTPUT_POINT VS_PointLighting(VS_INPUT IN)
{
	VS_OUTPUT_POINT OUT;

	float3 worldPos = mul(float4(IN.position, 1.0f), g_mWorld).xyz;
	float3 viewDir = g_mViewInv[3].xyz - worldPos;
	float3 lightDir = (light.pos - worldPos) / light.radius;
       
    float3 n = mul(IN.normal, (float3x3)g_mWorld);
	float3 t = mul(IN.tangent, (float3x3)g_mWorld);
	float3 b = cross(t, n);// * IN.tangent.w
	float3x3 tbnMatrix = float3x3(t.x, b.x, n.x,
	                              t.y, b.y, n.y,
	                              t.z, b.z, n.z);
			
	OUT.position = mul(float4(IN.position, 1.0f), g_mWorldViewProjection);
	OUT.texCoord = IN.texCoord;
	OUT.viewDir = mul(viewDir, tbnMatrix);
	OUT.lightDir = mul(lightDir, tbnMatrix);
	OUT.WorldNormal = mul(IN.normal, (float3x3)g_mWorld);//////////////////////////
	OUT.diffuse = material.diffuse * light.diffuse;
	OUT.specular = material.specular * light.specular;
	
	return OUT;
}

VS_OUTPUT_SPOT VS_SpotLighting(VS_INPUT IN)
{
    VS_OUTPUT_SPOT OUT;
    
    float3 worldPos = mul(float4(IN.position, 1.0f), g_mWorld).xyz;
    float3 viewDir = g_mViewInv[3].xyz - worldPos;
	float3 lightDir = (light.pos - worldPos) / light.radius;
    
    float3 n = mul(IN.normal, (float3x3)g_mWorldInverseTranspose);
	float3 t = mul(IN.tangent.xyz, (float3x3)g_mWorldInverseTranspose);
	float3 b = cross(t, n);
	float3x3 tbnMatrix = float3x3(t.x, b.x, n.x,
	                              t.y, b.y, n.y,
	                              t.z, b.z, n.z);
		       
    OUT.position = mul(float4(IN.position, 1.0f), g_mWorldViewProjection);
	OUT.texCoord = IN.texCoord;
	OUT.viewDir = mul(viewDir, tbnMatrix);
	OUT.lightDir = mul(lightDir, tbnMatrix);
    OUT.spotDir = mul(light.dir, tbnMatrix);
    OUT.diffuse = material.diffuse * light.diffuse;
	OUT.specular = material.specular * light.specular;
       
    return OUT;
}

//-----------------------------------------------------------------------------
// Pixel Shaders.
//-----------------------------------------------------------------------------

float4 PS_DirLighting(VS_OUTPUT_DIR IN) : COLOR
{
    float3 n = normalize(tex2D(normalMap, IN.texCoord).rgb * 2.0f - 1.0f);
    float3 h = normalize(IN.halfVector);
    float3 l = normalize(IN.lightDir);
	
	float3 v = normalize(IN.viewDir);
	//float3 h2 = normalize(float3(l.x, l.y, -l.z) + v);
	
	float4 cBaseColor = tex2D( colorMap, IN.texCoord );//////////////////////////////////////////////
	IN.WorldNormal = float3(IN.WorldNormal.x, IN.WorldNormal.z, IN.WorldNormal.y);
	float3 reflVect = reflect( float3(v.x, v.y, v.z), float3(n.x, n.y, n.z));//float3(n.x, n.z, n.y)
	float4 cEnvColor = texCUBE( envMap, reflVect );
	float4 cIrradColor = texCUBE( irradMap, IN.WorldNormal);

    float nDotL = saturate(dot(n, l));
	//float nDotL2 = saturate(dot(n, float3(l.x, l.y, -l.z)));
    float nDotH = saturate(dot(n, h));
	//float nDotH2 = saturate(dot(n, h2));
    //float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);
	//float power2 = (nDotL2 == 0.0f) ? 0.0f : pow(nDotH2, 50);
	float power3 = (nDotL == 0.0f) ? 0.0f : pow(nDotH, 400);
    
	//float4 color = (cIrradColor)+(blue* nDotL2) + (orange * nDotL) + (power2) + (IN.specular * power * specPower * cBaseColor.w);/////////////////////
	float4 ColorBase = cIrradColor*0.5 + (cIrradColor*0.5 * nDotL) + (cIrradColor*0.5 * specPower*power3);
	float4 ColorSeats = orange + (orange * nDotL) + (orange * specPower*power3);
	float4 ColorMetal = cEnvColor*white + (cEnvColor*white * nDotL) + (cEnvColor*white * specPower*power3);
	float4 baseColor = {0.0f, 0.0f, 0.0f, 0.0f};
	if (g_outNormals)
	    baseColor = float4(normalize( IN.WorldNormal ) * 0.5 + 0.5, 1.0f);
	else
	{
	    if (g_mode)
	        baseColor = lerp(lerp(ColorBase, ColorSeats, cBaseColor.b), ColorMetal, cBaseColor.g)*cBaseColor.r;
	    else
	        baseColor = cIrradColor;
		baseColor.w = 1.0f;
	}
	//float4 color = (material.ambient * (globalAmbient + light.ambient)) + (IN.diffuse * nDotL) + (IN.specular * power * specPower * cBaseColor.w);
	return baseColor;//color * cBaseColor;
}

float4 PS_PointLighting(VS_OUTPUT_POINT IN) : COLOR
{
    float atten = saturate(1.0f - dot(IN.lightDir, IN.lightDir));

	float3 n = normalize(tex2D(normalMap, IN.texCoord).rgb * 2.0f - 1.0f);
    float3 l = normalize(IN.lightDir);
    float3 v = normalize(IN.viewDir);
    float3 h = normalize(l + v);
	float4 cBaseColor = tex2D( colorMap, IN.texCoord );
	
	IN.WorldNormal = float3(IN.WorldNormal.x, IN.WorldNormal.z, IN.WorldNormal.y);
	float4 cIrradColor = texCUBE( irradMap, IN.WorldNormal);
    
    float nDotL = saturate(dot(n, l));
    float nDotH = saturate(dot(n, h));
    float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);

    float4 color = (atten * cIrradColor) + (IN.diffuse * nDotL * atten) + (IN.specular * power * specPower * atten * cBaseColor.w);
                   
	//float3 reflVect = -reflect( v, n );
	//float3 cEnvColor = texCUBE( envMap, l ).rgb;
	return color * cBaseColor;// tex2D(colorMap, IN.texCoord);
}

float4 PS_SpotLighting(VS_OUTPUT_SPOT IN) : COLOR
{
    float atten = saturate(1.0f - dot(IN.lightDir, IN.lightDir));
    
	float3 l = normalize(IN.lightDir);
    float2 cosAngles = cos(float2(light.spotOuterCone, light.spotInnerCone) * 0.5f);
    float spotDot = dot(-l, normalize(IN.spotDir));
    float spotEffect = smoothstep(cosAngles[0], cosAngles[1], spotDot);
    
    atten *= spotEffect;

    float3 n = normalize(tex2D(normalMap, IN.texCoord).rgb * 2.0f - 1.0f);
	float3 v = normalize(IN.viewDir);
	float3 h = normalize(l + v);
    
    float nDotL = saturate(dot(n, l));
    float nDotH = saturate(dot(n, h));
    float power = (nDotL == 0.0f) ? 0.0f : pow(nDotH, material.shininess);
    
    float4 color = (material.ambient * (globalAmbient + (atten * light.ambient))) +
                   (IN.diffuse * nDotL * atten) + (IN.specular  * specPower * power * atten);
    
	return color * tex2D(colorMap, IN.texCoord);
}

//-----------------------------------------------------------------------------
// Techniques.
//-----------------------------------------------------------------------------

technique NormalMappingDirectionalLighting
{
	pass
	{
		VertexShader = compile vs_2_0 VS_DirLighting();
		PixelShader = compile ps_2_0 PS_DirLighting();
	}
}

technique NormalMappingPointLighting
{
    pass
    {
        VertexShader = compile vs_2_0 VS_PointLighting();
        PixelShader = compile ps_2_0 PS_PointLighting();
    }
}

technique NormalMappingSpotLighting
{
    pass
    {
        VertexShader = compile vs_2_0 VS_SpotLighting();
        PixelShader = compile ps_2_0 PS_SpotLighting();
    }
}
